-------------------------------------------------------------------------------
-- TDC Core / CERN
-------------------------------------------------------------------------------
--
-- unit name: tdc_channel
--
-- author: Sebastien Bourdeauducq, sebastien@milkymist.org
--
-- description: Per-channel processing
--
-- references: http://www.ohwr.org/projects/tdc-core
--
-------------------------------------------------------------------------------
-- last changes:
-- 2011-11-07 SB Pre-inversion
-- 2011-10-25 SB Disable ring oscillator on reset
-- 2011-08-03 SB Created file
-------------------------------------------------------------------------------

-- Copyright (C) 2011 CERN
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU Lesser General Public License as published by
-- the Free Software Foundation, version 3 of the License.
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
-- You should have received a copy of the GNU Lesser General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.

-- DESCRIPTION:
-- This contains the elements needed for each channel:
--  * Delay line
--  * Encoder
--  * LUT
--  * Deskew stage
--  * Online calibration ring oscillator

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.tdc_package.all;
use work.genram_pkg.all;
use work.wishbone_pkg.all;
use work.tdc_wbgen2_pkg.all;
use work.gencores_pkg.all;

entity tdc_channel is
  generic(
    -- Number of CARRY4 elements.
    g_CARRY4_COUNT : positive := 124;
    -- Number of raw output bits.
    g_RAW_COUNT    : positive := 9;
    -- Number of fractional part bits.
    g_FP_COUNT     : positive := 13;
    -- Number of coarse counter bits.
    g_COARSE_COUNT : positive := 25;
    -- Length of the ring oscillator.
    g_RO_LENGTH    : positive := 31
    );
  port(
    clk_sys_i   : in std_logic;
    rst_sys_n_i : in std_logic;

    clk_tdc_i   : in std_logic;
    rst_tdc_n_i : in std_logic;

    clk_cal_i : in std_logic;

    -- Coarse counter and deskew inputs.
    coarse_i : in std_logic_vector(g_COARSE_COUNT-1 downto 0);

    -- Signal input.
    signal_i : in std_logic;

    slave_i : in  t_wishbone_slave_in;
    slave_o : out t_wishbone_slave_out

--    calib_sel_i : in std_logic;

    -- Detection outputs.
    --detect_o   : out std_logic;
    --polarity_o : out std_logic;
    --raw_o      : out std_logic_vector(g_RAW_COUNT-1 downto 0);
    --fp_o       : out std_logic_vector((g_COARSE_COUNT+g_FP_COUNT)-1 downto 0);
    --raw_taps_o : out std_logic_vector(g_CARRY4_COUNT*4- 1 downto 0);

    -- LUT access.
    --lut_a_i  : in  std_logic_vector(g_RAW_COUNT-1 downto 0);
    --lut_we_i : in  std_logic;
    --lut_d_i  : in  std_logic_vector(g_FP_COUNT-1 downto 0);
    --lut_d_o  : out std_logic_vector(g_FP_COUNT-1 downto 0);

   -- Calibration ring oscillator.
--  ro_en_i  : in  std_logic;
--    ro_clk_o : out std_logic
    );
end entity;

architecture rtl of tdc_channel is
  signal calib_sel_d  : std_logic;
  signal muxed_signal : std_logic;

  signal tdc_inv_input_signal   : std_logic;

  attribute keep : string;
  attribute keep of tdc_inv_input_signal : signal is "true";

  signal taps         : std_logic_vector(4*g_CARRY4_COUNT-1 downto 0);
  signal taps_latched : std_logic_vector(4*g_CARRY4_COUNT-1 downto 0);
  signal ipolarity    : std_logic;
  signal polarity     : std_logic;
  signal polarity_d1  : std_logic;
  signal polarity_d2  : std_logic;
  signal detect_d1    : std_logic;
  signal raw, raw_d1  : std_logic_vector(g_RAW_COUNT-1 downto 0);
  signal raw_d2       : std_logic_vector(g_RAW_COUNT-1 downto 0);
  signal lut_rvalue   : std_logic_vector(g_FP_COUNT-1 downto 0);
  signal ro_en        : std_logic;

  signal tdc_slave_out : t_wishbone_slave_out;
  signal tdc_slave_in  : t_wishbone_slave_in;

  signal regs_in  : t_tdc_in_registers;
  signal regs_out : t_tdc_out_registers;

  signal calib_next_sample  : unsigned(14 downto 0);
  signal calib_cur_sample   : unsigned(14 downto 0);
  signal calib_step_d       : unsigned(14 downto 0);
  signal calib_offset_d       : unsigned(14 downto 0);
  signal calib_p            : std_logic;
  signal calib_rst_n        : std_logic;
  signal rst_tdc            : std_logic;
  signal ro_clk             : std_logic;
  signal detect             : std_logic;
  signal prev_taps_zero     : std_logic;
  signal taps_latched_ready : std_logic;

  constant c_max_tap_bank : integer := (g_CARRY4_COUNT*4 + 31) / 32;

  function f_max(x : integer; y : integer) return integer is
  begin
    if (x > y) then
      return x;
    else
      return y;
    end if;
  end f_max;

  function f_clamp(x : integer; max : integer) return integer is
  begin
    if (x > max) then
      return max;
    else
      return x;
    end if;
  end f_clamp;


begin

  rst_tdc <= not rst_tdc_n_i;

  --slave_o <= tdc_slave_out;
--  tdc_slave_in <= slave_i;

  U_Clock_Crossing : xwb_clock_crossing
    port map (
      slave_clk_i    => clk_sys_i,
      slave_rst_n_i  => rst_sys_n_i,
      slave_i        => slave_i,
      slave_o        => slave_o,
      master_clk_i   => clk_tdc_i,
      master_rst_n_i => rst_tdc_n_i,
      master_i       => tdc_slave_out,  
      master_o       => tdc_slave_in);

  U_WB_Slave : entity work.tdc_channel_wb
    port map (
      rst_n_i   => rst_tdc_n_i,
      clk_sys_i => clk_tdc_i,
      slave_i   => tdc_slave_in,
      slave_o   => tdc_slave_out,
      int_o     => open,
      regs_i    => regs_in,
      regs_o    => regs_out);


  U_Sync_Calib_Sel : gc_sync_ffs
    port map (
      clk_i    => clk_cal_i,
      rst_n_i  => rst_tdc_n_i,
      data_i   => regs_out.calr_cal_mode_o,
      synced_o => calib_sel_d);

  U_Sync_Calib_Reset : gc_sync_ffs
    port map (
      clk_i    => clk_cal_i,
      rst_n_i  => '1',
      data_i   => rst_tdc_n_i,
      synced_o => calib_rst_n);


-- register calibration select signal to avoid glitches
  p_reg_calib_regs : process(clk_cal_i)
  begin
    if rising_edge(clk_cal_i) then
      if calib_rst_n = '0' then
        calib_step_d <= (others => '0');
        calib_offset_d <= (others => '0');
      else
        calib_step_d <= unsigned(regs_out.calr_cal_step_o);
        calib_offset_d <= unsigned(regs_out.calr_cal_offset_o);
      end if;
    end if;
  end process;

  p_calib_pattern_gen : process(clk_cal_i)
  begin
    if rising_edge(clk_cal_i) then
      if calib_rst_n = '0' then
        calib_cur_sample  <= (others => '0');
        calib_next_sample <= (others => '0');
        calib_p           <= '0';
      elsif calib_cur_sample = resize(calib_next_sample + calib_offset_d, calib_cur_sample'Length) and calib_step_d /= 0 then
        calib_next_sample <= calib_next_sample + calib_step_d;
        calib_cur_sample  <= calib_cur_sample + 1;
        calib_p           <= '1';
      elsif calib_step_d /= 0 then
        calib_cur_sample <= calib_cur_sample + 1;
        calib_p          <= '0';
      end if;
    end if;
  end process;

  muxed_signal <= calib_p when calib_sel_d = '1' else signal_i;
  tdc_inv_input_signal   <= muxed_signal xor not ipolarity;

  cmp_delayline : entity work.tdc_delayline
    generic map(
      g_WIDTH => g_CARRY4_COUNT
      )
    port map(
      clk_i    => clk_tdc_i,
      reset_i  => rst_tdc,
      signal_i => tdc_inv_input_signal,
      taps_o   => taps
      );

  cmp_lbc : entity work.tdc_lbc
    generic map(
      g_N      => g_RAW_COUNT,
      g_NIN    => g_CARRY4_COUNT*4,
      g_IGNORE => 2
      )
    port map(
      clk_i       => clk_tdc_i,
      reset_i     => rst_tdc,
      d_i         => taps,
      ipolarity_o => ipolarity,
      polarity_o  => polarity,
      count_o     => raw
      );

  process(clk_tdc_i)
  begin
    if rising_edge(clk_tdc_i) then
      if rst_tdc = '1' then
        prev_taps_zero     <= '1';
        taps_latched_ready <= '0';
      else
        if regs_out.raw_csr_ack_o = '1' then
          taps_latched_ready <= '0';
        end if;


        if unsigned(taps) = 0 then
          prev_taps_zero <= '1';
        elsif prev_taps_zero = '1' and unsigned(taps) /= 0 and taps_latched_ready = '0' then
          taps_latched       <= taps;
          taps_latched_ready <= '1';
          prev_taps_zero     <= '0';
        end if;

      end if;
    end if;
  end process;

  regs_in.raw_csr_valid_i <= taps_latched_ready;

  process(clk_tdc_i)
    variable idx : integer;
    variable idx_h, idx_l : integer;
  begin
    if rising_edge(clk_sys_i) then
      idx := to_integer(unsigned(regs_out.raw_bank_o));
      idx_l := idx * 32;
      idx_h := f_clamp(idx_l + 31, g_CARRY4_COUNT*4-1);
      regs_in.raw_taps_i <= std_logic_vector(resize(unsigned(taps_latched( idx_h downto idx_l)), 32));
      
    end if;
  end process;


  --cmp_lut : generic_dpram
  --  generic map(
  --    g_data_width               => g_FP_COUNT,
  --    g_size                     => 2**g_RAW_COUNT,
  --    g_with_byte_enable         => false,
  --    g_addr_conflict_resolution => "read_first",
  --    g_init_file                => "",
  --    g_dual_clock               => false
  --    )
  --  port map(
  --    clka_i => clk_i,
  --    clkb_i => '0',

  --    wea_i  => '0',
  --    bwea_i => (others => '0'),
  --    aa_i   => raw,
  --    da_i   => (others => '0'),
  --    qa_o   => lut_rvalue,

  --    web_i  => lut_we_i,
  --    bweb_i => (others => '0'),
  --    ab_i   => lut_a_i,
  --    db_i   => lut_d_i,
  --    qb_o   => lut_d_o
  --    );

  cmp_ringosc : entity work.tdc_ringosc
    generic map(
      g_LENGTH => g_RO_LENGTH
      )
    port map(
      en_i  => ro_en,
      clk_o => ro_clk
      );

  ro_en <= regs_out.csr_ro_en_o and rst_tdc_n_i;

  detect_d1 <= polarity_d1 xor polarity_d2;

  process(clk_tdc_i)
  begin
    if rising_edge(clk_tdc_i) then
      if rst_tdc = '1' then
        detect      <= '0';
        polarity_d1 <= '1';
        polarity_d2 <= '1';
        raw_d1      <= (others => '0');
        raw_d2      <= (others => '0');
      else
        detect      <= detect_d1;
        polarity_d1 <= polarity;
        raw_d1      <= raw;
        if detect_d1 = '1' then
          polarity_d2 <= polarity_d1;
          raw_d2      <= raw_d1;
        end if;
      end if;
    end if;
  end process;

--  polarity_o <= polarity_d2;
--  raw_o      <= raw_d2;

  -- Combine coarse counter value and deskew.
  --process(clk_i)
  --begin
  --  if rising_edge(clk_i) then
  --    if reset_i = '1' then
  --      fp_o <= (others => '0');
  --    else
  --      if detect_d1 = '1' then
  --        fp_o <= std_logic_vector(
  --          (resize(unsigned(coarse_i), g_FP_COUNT+g_COARSE_COUNT) sll g_FP_COUNT)
  --          - unsigned(lut_rvalue)
  --          + unsigned(deskew_i)
  --          );
  --      end if;
  --    end if;
  --  end if;
  --end process;

  regs_in.csr_valid_i <= '0';
  regs_in.csr_rdy_i <= '0';
  
end architecture;
