library ieee;

use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.wishbone_pkg.all;

entity tdc_core is

  generic (
    g_CARRY4_COUNT : positive := 124;
    g_RAW_COUNT    : positive := 9;
    g_FP_COUNT     : positive := 13;
    g_COARSE_COUNT : positive := 25;
    g_RO_LENGTH    : positive := 31);

  port (
    clk_sys_i   : in  std_logic;
    rst_sys_n_i : in  std_logic;
    clk_tdc_i   : in  std_logic;
    rst_tdc_n_i : in  std_logic;
    clk_cal_i   : in  std_logic;
    coarse_i    : in  std_logic_vector(g_COARSE_COUNT-1 downto 0) := (others => '0');
    signal_i    : in  std_logic;
    slave_i     : in  t_wishbone_slave_in;
    slave_o     : out t_wishbone_slave_out;
    dbg_txd_o   : out std_logic;
    dbg_rxd_i   : in  std_logic                                   := '1');

end entity tdc_core;

architecture rtl of tdc_core is

  constant c_cnx_slave_ports  : integer := 2;
  constant c_cnx_master_ports : integer := 3;

  constant c_master_host : integer := 0;
  constant c_master_urv  : integer := 1;

  constant c_slave_csr   : integer := 0;
  constant c_slave_vuart : integer := 1;
  constant c_slave_tdc   : integer := 2;

  signal cnx_master_in  : t_wishbone_master_in_array(c_cnx_master_ports-1 downto 0);
  signal cnx_master_out : t_wishbone_master_out_array(c_cnx_master_ports-1 downto 0);
  signal cnx_slave_in  : t_wishbone_slave_in_array(c_cnx_slave_ports-1 downto 0);
  signal cnx_slave_out : t_wishbone_slave_out_array(c_cnx_slave_ports-1 downto 0);

  constant c_cfg_base_addr : t_wishbone_address_array(c_cnx_master_ports-1 downto 0) :=
    (c_slave_csr   => x"00000000",
     c_slave_vuart => x"00001000",
     c_slave_tdc   => x"00002000"
     );

  constant c_cfg_base_mask : t_wishbone_address_array(c_cnx_master_ports-1 downto 0) :=
    (c_slave_csr   => x"0000f000",
     c_slave_vuart => x"0000f000",
     c_slave_tdc   => x"0000f000"
     );

begin

  slave_o <= cnx_slave_out(c_master_host);
  cnx_slave_in(c_master_host) <= slave_i;



  U_MCU : entity work.xwb_urv_mcs
    generic map (
      g_iram_size    => 65536,
      g_with_host_if => true
      )
    port map (
      clk_sys_i => clk_sys_i,
      rst_n_i   => rst_sys_n_i,
      host_wb_i => cnx_master_out(c_slave_csr),
      host_wb_o => cnx_master_in(c_slave_csr),
      dwb_o     => cnx_slave_in(c_master_urv),
      dwb_i     => cnx_slave_out(c_master_urv));

  U_Host_Intercon : xwb_crossbar
    generic map (
      g_num_masters => c_cnx_slave_ports,
      g_num_slaves  => c_cnx_master_ports,
      g_registered  => true,
      g_address     => c_cfg_base_addr,
      g_mask        => c_cfg_base_mask)
    port map (
      clk_sys_i => clk_sys_i,
      rst_n_i   => rst_sys_n_i,
      slave_i   => cnx_slave_in,
      slave_o   => cnx_slave_out,
      master_i  => cnx_master_in,
      master_o  => cnx_master_out);


  U_CH1 : entity work.tdc_channel
    generic map (
      g_CARRY4_COUNT => g_CARRY4_COUNT,
      g_RAW_COUNT    => g_RAW_COUNT,
      g_FP_COUNT     => g_FP_COUNT,
      g_COARSE_COUNT => g_COARSE_COUNT,
      g_RO_LENGTH    => g_RO_LENGTH)
    port map (
      clk_sys_i   => clk_sys_i,
      rst_sys_n_i => rst_sys_n_i,
      clk_tdc_i   => clk_tdc_i,
      rst_tdc_n_i => rst_tdc_n_i,
      clk_cal_i   => clk_cal_i,
      coarse_i    => coarse_i,
      signal_i    => signal_i,
      slave_i     => cnx_master_out(c_slave_tdc),
      slave_o     => cnx_master_in(c_slave_tdc));

  U_VUART : xwb_simple_uart
    generic map (
      g_with_virtual_uart   => true,
      g_with_physical_uart  => true,
      g_interface_mode      => PIPELINED,
      g_address_granularity => BYTE)
    port map (
      clk_sys_i  => clk_sys_i,
      rst_n_i    => rst_sys_n_i,
      slave_i    => cnx_master_out(c_slave_vuart),
      slave_o    => cnx_master_in(c_slave_vuart),
      uart_rxd_i => dbg_rxd_i,
      uart_txd_o => dbg_txd_o);
  
end rtl;

