-------------------------------------------------------------------------------
-- Title      : LM32 Embedded Microcontroller System (MCS)
-- Project    : General Cores Library
-------------------------------------------------------------------------------
-- File       : xwb_lm32_mcs.vhd
-- Author     : Tomasz Włostowski
-- Company    : CERN BE-CO-HT
-- Platform   : FPGA-generic
-- Standard   : VHDL'93
-------------------------------------------------------------------------------
-- Description: 
--
-- A small embedded microcontroller with some amount of RAM,
-- UART and a Wishbone bus for user peripherals. The code can be preloaded
-- or loaded on-the-fly throufh the Wishbone system bus.
-------------------------------------------------------------------------------
--
-- Copyright (c) 2014-2017 CERN
--
-- This source file is free software; you can redistribute it   
-- and/or modify it under the terms of the GNU Lesser General   
-- Public License as published by the Free Software Foundation; 
-- either version 2.1 of the License, or (at your option) any   
-- later version.                                               
--
-- This source is distributed in the hope that it will be       
-- useful, but WITHOUT ANY WARRANTY; without even the implied   
-- warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      
-- PURPOSE.  See the GNU Lesser General Public License for more 
-- details.                                                     
--
-- You should have received a copy of the GNU Lesser General    
-- Public License along with this source; if not, download it   
-- from http://www.gnu.org/licenses/lgpl-2.1.html
--
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.genram_pkg.all;
use work.wishbone_pkg.all;

entity xwb_urv_mcs is
  generic(
   -- size of the code/data memory in bytes
    g_iram_size        : integer := 65536;
    -- timeout (in clock cycles) for the Wishbone master.
    -- exceeding the timeout (e.g. due to an incorrect address/
    -- unresponsive slave) causes a WB error after g_bus_timeout cycles.
    g_bus_timeout      : integer := 30;
    -- file (.bram format) with the firmware to pre-load during synthesis.
    g_preload_firmware : string  := "";
    -- Enable host interface (allows loading code from a system bus)
    g_with_host_if     : boolean := true
    );

  port(
    clk_sys_i : in std_logic;
    rst_n_i   : in std_logic;
    irq_i     : in std_logic_vector(31 downto 0) := x"00000000";

    host_wb_i : in  t_wishbone_slave_in := cc_dummy_slave_in;
    host_wb_o : out t_wishbone_slave_out;

    dwb_o : out t_wishbone_master_out;
    dwb_i : in  t_wishbone_master_in
    );
end xwb_urv_mcs;

architecture wrapper of xwb_urv_mcs is

  constant c_address_bits : integer := 20;
  
  function f_x_to_zero (x : std_logic_vector) return std_logic_vector is
    variable tmp : std_logic_vector(x'length-1 downto 0);
  begin
    -- synthesis translate_off
    for i in 0 to x'length-1 loop
      if(x(i) = '1') then
        tmp(i) := '1';
      else
        tmp(i) := '0';
      end if;
    end loop;
    return tmp;
    -- synthesis translate_on
    return x;
  end function;

  constant c_iram_addr_width : integer := f_log2_size(g_iram_size)-2;

  signal cpu_reset, cpu_enable, cpu_reset_d : std_logic;

  signal bus_timeout     : unsigned(7 downto 0);
  signal bus_timeout_hit : std_logic;

  signal im_addr            : std_logic_vector(31 downto 0);
  signal im_data            : std_logic_vector(31 downto 0);
  signal im_valid           : std_logic;

  signal ha_im_addr                                : std_logic_vector(c_address_bits-1 downto 0);
  signal ha_im_wdata, ha_im_rdata                  : std_logic_vector(31 downto 0);
  signal ha_im_access, ha_im_write : std_logic;

  signal im_addr_muxed : std_logic_vector(c_address_bits-1 downto 0);

  signal dm_addr, dm_data_s, dm_data_l                            : std_logic_vector(31 downto 0);
  signal dm_data_select                                           : std_logic_vector(3 downto 0);
  signal dm_load, dm_store, dm_load_done, dm_store_done, dm_ready : std_logic;

  signal dm_cycle_in_progress, dm_is_wishbone : std_logic;

  signal dm_mem_rdata, dm_wb_rdata : std_logic_vector(31 downto 0);
  signal dm_wb_write, dm_select_wb : std_logic;
  signal dm_data_write             : std_logic;

  signal cpu_csr_udata_out, cpu_csr_uaddr_addr, cpu_csr_udata_in : std_logic_vector(31 downto 0);
  signal cpu_csr_udata_load                                      : std_logic;
  signal iram_bwe  : std_logic_vector(3 downto 0);
  
  function f_resize ( x :std_logic_vector; n : integer ) return std_logic_vector is
    variable rv : std_logic_vector(n-1 downto 0);
  begin
    rv := (others => '0');
    rv (x'length-1 downto 0) := x;
    return rv;
  end f_resize;

  signal cpu_dwb_out : t_wishbone_master_out;
  signal cpu_dwb_in : t_wishbone_master_in;

begin

  dm_is_wishbone <= dm_addr(c_address_bits-1);

    -- Wishbone bus arbitration / internal RAM access
  process(clk_sys_i)
  begin
    if rising_edge(clk_sys_i) then
      if(rst_n_i = '0') then
        cpu_dwb_out.cyc            <= '0';
        dm_cycle_in_progress <= '0';
        dm_load_done         <= '0';
        dm_store_done        <= '0';
        dm_select_wb         <= '0';
        bus_timeout <= (others => '0');
      else

        if(dm_cycle_in_progress = '0') then  -- access to internal memory
          bus_timeout <= (others => '0');
          
          if(dm_is_wishbone = '0') then
            if(dm_store = '1') then
              dm_load_done  <= '0';
              dm_store_done <= '1';
              dm_select_wb  <= '0';
            elsif (dm_load = '1') then
              dm_load_done  <= '1';
              dm_store_done <= '0';
              dm_select_wb  <= '0';
            else
              dm_store_done <= '0';
              dm_load_done  <= '0';
              dm_select_wb  <= '0';
            end if;
          else
            if(dm_load = '1' or dm_store = '1') then
              cpu_dwb_out.cyc   <= '1';
              cpu_dwb_out.stb   <= '1';
              cpu_dwb_out.we    <= dm_store;
              dm_wb_write <= dm_store;

              cpu_dwb_out.adr <= dm_addr;
              cpu_dwb_out.dat <= dm_data_s;
              cpu_dwb_out.sel <= dm_data_select;


              dm_load_done         <= '0';
              dm_store_done        <= '0';
              dm_cycle_in_progress <= '1';
            else
              dm_store_done        <= '0';
              dm_load_done         <= '0';
              dm_cycle_in_progress <= '0';
            end if;
          end if;
        else -- dm_cycle_in_progress = 1

          bus_timeout <= bus_timeout + 1;
          
          if(cpu_dwb_in.stall = '0') then
            cpu_dwb_out.stb <= '0';
          end if;

          if(cpu_dwb_in.ack = '1' or bus_timeout = g_bus_timeout) then
            if(dm_wb_write = '0') then
              dm_wb_rdata  <= cpu_dwb_in.dat;
              dm_select_wb <= '1';
              dm_load_done <= '1';
            else
              dm_store_done <= '1';
              dm_select_wb  <= '0';
            end if;

            dm_cycle_in_progress <= '0';
            cpu_dwb_out.cyc            <= '0';
          end if;
        end if;
      end if;
    end if;
  end process;

  dm_data_l     <= dm_wb_rdata when dm_select_wb = '1' else dm_mem_rdata;
  im_addr_muxed <= ha_im_addr  when ha_im_access = '1' else im_addr(c_address_bits-1 downto 0);
  dm_ready      <= '1';

  U_CPU_core : entity work.urv_cpu
    port map (
      clk_i            => clk_sys_i,
      rst_i            => cpu_reset,
      irq_i            => irq_i(0),
      im_addr_o        => im_addr,
      im_data_i        => im_data,
      im_valid_i       => im_valid,
      dm_addr_o        => dm_addr,
      dm_data_s_o      => dm_data_s,
      dm_data_l_i      => dm_data_l,
      dm_data_select_o => dm_data_select,
      dm_ready_i       => dm_ready,
      dm_store_o       => dm_store,
      dm_load_o        => dm_load,
      dm_load_done_i   => dm_load_done,
      dm_store_done_i  => dm_store_done,
      dbg_force_i => '0',
      dbg_insn_i => x"00000000",
      dbg_mbx_data_i => x"00000000",
      dbg_mbx_write_i => '0',
      dbg_insn_set_i => '0',
      dbg_tx_ready_i => '0',
      dbg_rx_data_i => x"00",
      dbg_rx_valid_i => '0' );

    dm_data_write <= not dm_is_wishbone and dm_store;

    gen_iram_blocks : for i in 0 to 3 generate
    
    iram : generic_dpram
      generic map (
        g_data_width               => 8,
        g_size                     => g_iram_size / 4,
        g_with_byte_enable         => false,
        g_dual_clock               => false,
        g_addr_conflict_resolution => "dont_care")
      port map (
        rst_n_i => rst_n_i,
        clka_i  => clk_sys_i,

        wea_i => ha_im_write,
        aa_i  => im_addr_muxed(f_log2_size(g_iram_size)-1 downto 2),
        da_i  => ha_im_wdata(8*i+7 downto 8*i),
        qa_o  => im_data(8*i+7 downto 8*i),

        clkb_i => clk_sys_i,
        web_i  => iram_bwe(i),
        ab_i   => dm_addr(f_log2_size(g_iram_size)-1 downto 2),
        db_i   => dm_data_s(8*i+7 downto 8*i),
        qb_o   => dm_mem_rdata(8*i+7 downto 8*i)
        );

    iram_bwe(i) <= dm_data_select(i) and dm_data_write;

    end generate gen_iram_blocks;

    -- U_iram : entity work.urv_iram
    -- generic map (
    --   g_size       => g_iram_size,
    --   g_init_file  => g_preload_firmware,
    --   g_simulation => 0)
    -- port map (
    --   clk_i => clk_sys_i,
    --   ena_i  => '1',
    --   wea_i  => ha_im_write,
    --   bwea_i => "1111",
    --   aa_i   => x"deadbeef", --x"000" & im_addr_muxed,
    --   da_i   => ha_im_wdata,
    --   qa_o   => im_data,
    --   enb_i  => '1',
    --   bweb_i => dm_data_select,
    --   web_i  => dm_data_write,
    --   ab_i   => dm_addr,
    --   db_i   => dm_data_s,
    --   qb_o   => dm_mem_rdata
    --   );

  ha_im_rdata <= im_data;

  process(clk_sys_i)
  begin
    if rising_edge(clk_sys_i) then
      if(cpu_reset = '1') then
        im_valid  <= '0';
        cpu_reset_d <= '1';
      else
        cpu_reset_d <= cpu_reset;
        im_valid  <= not ha_im_access and (not cpu_reset_d);
      end if;
    end if;
  end process;


  dwb_o <= cpu_dwb_out;
  cpu_dwb_in.dat <= f_x_to_zero(dwb_i.dat);
  cpu_dwb_in.ack <= dwb_i.ack;
  cpu_dwb_in.err <= dwb_i.err;
  cpu_dwb_in.rty <= dwb_i.rty;
  cpu_dwb_in.stall <= dwb_i.stall;
  
  
  cpu_reset   <= not rst_n_i or (not cpu_enable);
  ha_im_access <= not cpu_enable;
  
  p_local_regs : process(clk_sys_i)
  begin
    if rising_edge(clk_sys_i) then
      if rst_n_i = '0' then
        cpu_enable           <= '0';
        host_wb_o.ack   <= '0';
        host_wb_o.err   <= '0';
        host_wb_o.rty   <= '0';
        host_wb_o.stall <= '0';
        
      else
        ha_im_write <= '0';
        host_wb_o.ack <= '0';

        if host_wb_i.cyc = '1' and host_wb_i.stb = '1' then

          host_wb_o.ack <= '1';

          if host_wb_i.we = '1' then
            case host_wb_i.adr(3 downto 0) is
              when "0000" =>            -- csr
                cpu_enable <= host_wb_i.dat(0);

              when "0100" =>            -- data
                ha_im_wdata <= host_wb_i.dat;
                ha_im_write <= '1';
                
              when "1000" =>            -- addr
                ha_im_addr <= host_wb_i.dat(c_address_bits-3 downto 0) & "00" ;
              when others => null;
            end case;
          else
            case host_wb_i.adr(3 downto 0) is
              when "0000" =>
                host_wb_o.dat(0)           <= cpu_enable;
                host_wb_o.dat(31 downto 1) <= std_logic_vector(to_unsigned(g_iram_size, 31));
              when "0100" =>
                host_wb_o.dat <= ha_im_rdata;
              when others => null;
            end case;
          end if;
        end if;
      end if;
    end if;
  end process;
  

end wrapper;
