library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;
use work.gn4124_core_pkg.all;
use work.gencores_pkg.all;
use work.wishbone_pkg.all;

library unisim;
use unisim.vcomponents.all;

entity spec_top is
  generic
    (g_simulation        : boolean := false;
     g_sim_bypass_gennum : boolean := false
     );
                                         -- when instantiated in a test-bench
  port(
    clk_125m_pllref_p_i : in std_logic;  -- 125 MHz PLL reference
    clk_125m_pllref_n_i : in std_logic;
    clk_20m_vcxo_i      : in std_logic;  -- 20 MHz VCXO

    uart_rxd_i : in  std_logic := '1';
    uart_txd_o : out std_logic;

    ------------------------------------------------------------------------
    -- GN4124 PCI bridge pins
    ------------------------------------------------------------------------

    gn_rst_n      : in  std_logic;      -- reset from gn4124 (rstout18_n)
    -- general purpose interface
    gn_gpio       : out std_logic_vector(1 downto 0);  -- gpio[0] -> gn4124 gpio8
    -- pcie to local [inbound data] - rx
    gn_p2l_rdy    : out std_logic;      -- rx buffer full flag
    gn_p2l_clkn   : in  std_logic;      -- receiver source synchronous clock-
    gn_p2l_clkp   : in  std_logic;      -- receiver source synchronous clock+
    gn_p2l_data   : in  std_logic_vector(15 downto 0);  -- parallel receive data
    gn_p2l_dframe : in  std_logic;      -- receive frame
    gn_p2l_valid  : in  std_logic;      -- receive data valid
    -- inbound buffer request/status
    gn_p_wr_req   : in  std_logic_vector(1 downto 0);  -- pcie write request
    gn_p_wr_rdy   : out std_logic_vector(1 downto 0);  -- pcie write ready
    gn_rx_error   : out std_logic;      -- receive error
    -- local to parallel [outbound data] - tx
    gn_l2p_data   : out std_logic_vector(15 downto 0);  -- parallel transmit data
    gn_l2p_dframe : out std_logic;      -- transmit data frame
    gn_l2p_valid  : out std_logic;      -- transmit data valid
    gn_l2p_clkn   : out std_logic;  -- transmitter source synchronous clock-
    gn_l2p_clkp   : out std_logic;  -- transmitter source synchronous clock+
    gn_l2p_edb    : out std_logic;      -- packet termination and discard
    -- outbound buffer status
    gn_l2p_rdy    : in  std_logic;      -- tx buffer full flag
    gn_l_wr_rdy   : in  std_logic_vector(1 downto 0);  -- local-to-pcie write
    gn_p_rd_d_rdy : in  std_logic_vector(1 downto 0);  -- pcie-to-local read response data ready
    gn_tx_error   : in  std_logic;      -- transmit error
    gn_vc_rdy     : in  std_logic_vector(1 downto 0)  -- channel ready


    -- Bypass GN4124 core, useful only in simulation
    -- Feed fake timestamps bypassing acam - used only in simulation
    -- synthesis translate_off
;
    sim_wb_i : in  t_wishbone_slave_in := cc_dummy_slave_in;
    sim_wb_o : out t_wishbone_slave_out

-- synthesis translate_on

    );

end spec_top;

architecture rtl of spec_top is



  function f_bool2int (x : boolean) return integer is
  begin
    if(x) then
      return 1;
    else
      return 0;
    end if;
  end f_bool2int;

  -- Master ports on the wishbone crossbar
  constant c_NUM_WB_MASTERS : integer := 1;
  constant c_WB_SLAVE_TDC   : integer := 0;

  -- Slave port on the wishbone crossbar
  constant c_NUM_WB_SLAVES : integer := 1;
  constant c_MASTER_GENNUM : integer := 0;


  constant c_CNX_BASE_ADDR : t_wishbone_address_array(0 to 0) :=
    (c_WB_SLAVE_TDC => x"00000000");

  constant c_CNX_BASE_MASK : t_wishbone_address_array(0 to 0) :=
    (c_WB_SLAVE_TDC => x"000F0000");

---------------------------------------------------------------------------------------------------
--                                            Signals                                            --
---------------------------------------------------------------------------------------------------
  -- Clocks and resets
  signal clk_sys_62m5   : std_logic;
  signal rst_sys_62m5_n : std_logic;

  signal cnx_master_out : t_wishbone_master_out_array(c_NUM_WB_MASTERS-1 downto 0);
  signal cnx_master_in  : t_wishbone_master_in_array(c_NUM_WB_MASTERS-1 downto 0);
  signal cnx_slave_out  : t_wishbone_slave_out_array(c_NUM_WB_SLAVES-1 downto 0);
  signal cnx_slave_in   : t_wishbone_slave_in_array(c_NUM_WB_SLAVES-1 downto 0);
  signal gn_wb_adr      : std_logic_vector(31 downto 0);

  function f_to_string(x : boolean) return string is
  begin
    if x then
      return "TRUE";
    else
      return "FALSE";
    end if;
  end f_to_string;

  signal pll_arst                   : std_logic := '0';
  signal clk_125m_pllref_buf        : std_logic;
  signal clk_sys_prebuf, clk_sys_fb : std_logic;
  signal clk_ref_prebuf, clk_ref : std_logic;

  signal clk_cal1_fb, clk_cal2_fb, clk_cal1, clk_cal2, clk_cal1_prebuf, clk_cal2_prebuf : std_logic;
  signal pll_sys_locked : std_logic;
  signal gn4124_status : std_logic_vector(31 downto 0);

  signal rst_ref_n : std_logic;
  signal clk_cal : std_logic;
  
--=================================================================================================
--                                       architecture begin
--=================================================================================================
begin

  U_Reset_Gen: entity work.spec_reset_gen
    port map (
      clk_sys_i        => clk_sys_62m5,
      rst_pcie_n_a_i   => gn_rst_n,
      rst_button_n_a_i => '1',
      rst_n_o          => rst_sys_62m5_n);
  
  -- active high async reset for PLLs
  pll_arst <= not gn_rst_n;

  cmp_sys_clk_pll : PLL_BASE
    generic map (
      BANDWIDTH          => "OPTIMIZED",
      CLK_FEEDBACK       => "CLKFBOUT",
      COMPENSATION       => "INTERNAL",
      DIVCLK_DIVIDE      => 1,
      CLKFBOUT_MULT      => 8,
      CLKFBOUT_PHASE     => 0.000,
      CLKOUT0_DIVIDE     => 16,
      CLKOUT0_PHASE      => 0.000,
      CLKOUT0_DUTY_CYCLE => 0.500,
      CLKOUT1_DIVIDE     => 8,
      CLKOUT1_PHASE      => 0.000,
      CLKOUT1_DUTY_CYCLE => 0.500,
      CLKIN_PERIOD       => 8.0,
      REF_JITTER         => 0.016)
    port map (
      CLKFBOUT => clk_sys_fb,
      CLKOUT0  => clk_sys_prebuf,
      CLKOUT1  => clk_ref_prebuf,
      LOCKED   => pll_sys_locked,
      RST      => pll_arst,
      CLKFBIN  => clk_sys_fb,
      CLKIN    => clk_125m_pllref_buf);

  -- System PLL input clock buffer
  cmp_clk_sys_buf_i : IBUFGDS
    port map (
      O  => clk_125m_pllref_buf,
      I  => clk_125m_pllref_p_i,
      IB => clk_125m_pllref_n_i);

  -- System PLL output clock buffer
  cmp_clk_sys_buf_o : BUFG
    port map (
      O => clk_sys_62m5,
      I => clk_sys_prebuf);
  
--  System PLL output clock buffer
  cmp_clk_ref_buf_o : BUFG
    port map (
      O => clk_ref,
      I => clk_ref_prebuf);

  cmp_cal_clk_pll_1 : PLL_BASE
    generic map (
      BANDWIDTH          => "OPTIMIZED",
      CLK_FEEDBACK       => "CLKFBOUT",
      COMPENSATION       => "INTERNAL",
      DIVCLK_DIVIDE      => 4,
      CLKFBOUT_MULT      => 31,
      CLKFBOUT_PHASE     => 0.000,
      CLKOUT0_DIVIDE     => 32,
      CLKOUT0_PHASE      => 0.000,
      CLKOUT0_DUTY_CYCLE => 0.500,
      CLKIN_PERIOD       => 8.0,
      REF_JITTER         => 0.016)
    port map (
      CLKFBOUT => clk_cal1_fb,
      CLKOUT0  => clk_cal1_prebuf,
      RST      => pll_arst,
      CLKFBIN  => clk_cal1_fb,
      CLKIN    => clk_125m_pllref_buf);


  cmp_clk_cal1_buf : BUFG
    port map (
      O => clk_cal1,
      I => clk_cal1_prebuf);

  cmp_cal_clk_pll_2 : PLL_BASE
    generic map (
      BANDWIDTH          => "OPTIMIZED",
      CLK_FEEDBACK       => "CLKFBOUT",
      COMPENSATION       => "INTERNAL",
      DIVCLK_DIVIDE      => 1,
      CLKFBOUT_MULT      => 33,
      CLKFBOUT_PHASE     => 0.000,
      CLKOUT0_DIVIDE     => 16,
      CLKOUT0_PHASE      => 0.000,
      CLKOUT0_DUTY_CYCLE => 0.500,
      CLKIN_PERIOD       => 33.03,
      REF_JITTER         => 0.016)
    port map (
      CLKFBOUT => clk_cal2_fb,
      CLKOUT0  => clk_cal2_prebuf,
      RST      => pll_arst,
      CLKFBIN  => clk_cal2_fb,
      CLKIN    => clk_cal1);

--  System PLL output clock buffer
  cmp_clk_cal2_buf : BUFG
    port map (
      O => clk_cal,
      I => clk_cal2_prebuf);
  
  cmp_sdb_crossbar : xwb_crossbar
    generic map
    (g_num_masters => c_NUM_WB_SLAVES,
     g_num_slaves  => c_NUM_WB_MASTERS,
     g_registered  => true,
     g_address     => c_CNX_BASE_ADDR,
     g_mask        => c_CNX_BASE_MASK)
    port map
    (clk_sys_i => clk_sys_62m5,
     rst_n_i   => rst_sys_62m5_n,
     slave_i   => cnx_slave_in,
     slave_o   => cnx_slave_out,
     master_i  => cnx_master_in,
     master_o  => cnx_master_out);


---------------------------------------------------------------------------------------------------
--                                           GN4124 CORE                                         --
---------------------------------------------------------------------------------------------------

  gen_with_gennum : if g_sim_bypass_gennum = false generate
    cmp_gn4124_core : gn4124_core
      port map
      (rst_n_a_i    => gn_rst_n,
       status_o     => gn4124_status,
       ---------------------------------------------------------
       -- P2L Direction
       --
       -- Source Sync DDR related signals
       p2l_clk_p_i  => gn_p2l_clkp,
       p2l_clk_n_i  => gn_p2l_clkn,
       p2l_data_i   => gn_p2l_data,
       p2l_dframe_i => gn_p2l_dframe,
       p2l_valid_i  => gn_p2l_valid,
       -- P2L Control
       p2l_rdy_o    => gn_p2l_rdy,
       p_wr_req_i   => gn_p_wr_req,
       p_wr_rdy_o   => gn_p_wr_rdy,
       rx_error_o   => gn_rx_error,
       vc_rdy_i     => gn_vc_rdy,

       ---------------------------------------------------------
       -- L2P Direction
       --
       -- Source Sync DDR related signals
       l2p_clk_p_o  => gn_l2p_clkp,
       l2p_clk_n_o  => gn_l2p_clkn,
       l2p_data_o   => gn_l2p_data,
       l2p_dframe_o => gn_l2p_dframe,
       l2p_valid_o  => gn_l2p_valid,
       -- L2P Control
       l2p_edb_o    => gn_l2p_edb,
       l2p_rdy_i    => gn_l2p_rdy,
       l_wr_rdy_i   => gn_l_wr_rdy,
       p_rd_d_rdy_i => gn_p_rd_d_rdy,
       tx_error_i   => gn_tx_error,

       irq_p_i   => '0',
       irq_p_o   => open,

       -- CSR WISHBONE interface (master pipelined)
       csr_clk_i   => clk_sys_62m5,
       csr_adr_o   => gn_wb_adr,
       csr_dat_o   => cnx_slave_in(c_MASTER_GENNUM).dat,
       csr_sel_o   => cnx_slave_in(c_MASTER_GENNUM).sel,
       csr_stb_o   => cnx_slave_in(c_MASTER_GENNUM).stb,
       csr_we_o    => cnx_slave_in(c_MASTER_GENNUM).we,
       csr_cyc_o   => cnx_slave_in(c_MASTER_GENNUM).cyc,
       csr_dat_i   => cnx_slave_out(c_MASTER_GENNUM).dat,
       csr_ack_i   => cnx_slave_out(c_MASTER_GENNUM).ack,
       csr_stall_i => cnx_slave_out(c_MASTER_GENNUM).stall,
       csr_err_i   => '0',
       csr_rty_i   => '0'
       );

    --  --  --  --  --  --  --  --  --  --  --  --  --  --  --  --  --  --
    -- Convert 32-bit word address into byte address for crossbar
    cnx_slave_in(c_MASTER_GENNUM).adr <= gn_wb_adr(29 downto 0) & "00";
    gn_gpio <= (others => '0');
    
  end generate gen_with_gennum;

  gen_without_gennum : if g_sim_bypass_gennum generate
    -- synthesis translate_off
    cnx_slave_in(c_MASTER_GENNUM) <= sim_wb_i;
    sim_wb_o                      <= cnx_slave_out(c_MASTER_GENNUM);
  -- synthesis translate_on
  end generate gen_without_gennum;


  U_Sync_Ref_Reset : gc_sync_ffs
    port map (
      clk_i    => clk_ref,
      rst_n_i  => '1',
      data_i   => rst_sys_62m5_n,
      synced_o => rst_ref_n);
  
  U_TDC_Core: entity work.tdc_core
    port map (
      clk_sys_i   => clk_sys_62m5,
      rst_sys_n_i => rst_sys_62m5_n,
      clk_tdc_i   => clk_ref,
      rst_tdc_n_i => rst_ref_n,

      clk_cal_i   => clk_cal,
      signal_i    => '0',
      slave_i     => cnx_master_out(c_WB_SLAVE_TDC),
      slave_o     => cnx_master_in(c_WB_SLAVE_TDC),
      dbg_txd_o   => uart_txd_o,
      dbg_rxd_i   => uart_rxd_i);


end rtl;
